from typing import Callable
from gaussian_splatting import GaussianModel

from gaussian_splatting.trainer.densifier import AbstractDensifier, NoopDensifier, OpacityPruner
from .trainer import SplitClonePartialDensifierTrainerWrapper


def PartialDensificationTrainerWrapper(
        noargs_base_densifier_constructor: Callable[[GaussianModel, float], AbstractDensifier],
        model: GaussianModel,
        scene_extent: float,
        *args,
        prune_from_iter=1000,
        prune_until_iter=15000,
        prune_interval=100,
        prune_screensize_threshold=20,
        prune_percent_too_big=1,
        prune_opacity_threshold=0.005,
        **kwargs):
    # https://github.com/yindaheng98/gaussian-splatting/blob/56576b647d9c5bd05300f5640cd03a8c75a760bc/gaussian_splatting/trainer/densifier/combinations.py#L10
    return SplitClonePartialDensifierTrainerWrapper(
        lambda model, scene_extent: OpacityPruner(
            noargs_base_densifier_constructor(model, scene_extent),
            scene_extent,
            prune_from_iter=prune_from_iter,
            prune_until_iter=prune_until_iter,
            prune_interval=prune_interval,
            prune_screensize_threshold=prune_screensize_threshold,
            prune_percent_too_big=prune_percent_too_big,
            prune_opacity_threshold=prune_opacity_threshold
        ),
        model, scene_extent,
        *args, **kwargs
    )


def PartialDensificationTrainer(
        model: GaussianModel,
        scene_extent: float,
        *args, **kwargs):
    # https://github.com/yindaheng98/gaussian-splatting/blob/56576b647d9c5bd05300f5640cd03a8c75a760bc/gaussian_splatting/trainer/densifier/combinations.py#L38
    return PartialDensificationTrainerWrapper(
        lambda model, scene_extent: NoopDensifier(model),
        model,
        scene_extent,
        *args, **kwargs
    )



from gaussian_splatting import GaussianModel
from gaussian_splatting.dataset import CameraDataset
from gaussian_splatting.trainer import AbstractTrainer
from gaussian_splatting.trainer.extensions import ScaleRegularizeTrainerWrapper
from lapisgs.dataset import RescaleFixedTrainableCameraDataset, RescaleColmapCameraDataset, RescaleTrainableCameraDataset, RescaleColmapTrainableCameraDataset
from lapisgs.trainer import Trainer, CameraTrainer, DepthTrainer, DepthCameraTrainer


def prepare_dataset(source: str, device: str, trainable_camera: bool = False, load_camera: str = None, load_depth=False, rescale_factor=1.0) -> CameraDataset:
    if trainable_camera:
        dataset = (RescaleTrainableCameraDataset.from_json(load_camera, load_depth=load_depth, rescale_factor=rescale_factor)
                   if load_camera else RescaleColmapTrainableCameraDataset(source, load_depth=load_depth, rescale_factor=rescale_factor)).to(device)
    else:
        dataset = (RescaleFixedTrainableCameraDataset(load_camera, load_depth=load_depth, rescale_factor=rescale_factor)
                   if load_camera else RescaleColmapCameraDataset(source, load_depth=load_depth, rescale_factor=rescale_factor)).to(device)
    return dataset


modes = {
    "base": DepthTrainer,
    "camera": DepthCameraTrainer,
    "nodepth-base": Trainer,
    "nodepth-camera": CameraTrainer,
}


def prepare_trainer(gaussians: GaussianModel, dataset: CameraDataset, mode: str, trainable_camera: bool = False, load_ply: str = None, with_scale_reg=False, configs={}) -> AbstractTrainer:
    if not load_ply:
        from gaussian_splatting.prepare import prepare_trainer as legacy_prepare_trainer
        modemap = {
            "base": "densify",
            "camera": "camera-densify",
            "nodepth-base": "nodepth-densify",
            "nodepth-camera": "nodepth-camera-densify",
        }
        return legacy_prepare_trainer(
            gaussians=gaussians, dataset=dataset, mode=modemap[mode], trainable_camera=trainable_camera, load_ply=load_ply, with_scale_reg=with_scale_reg, configs=configs)
    constructor = modes[mode]
    if with_scale_reg:
        constructor = lambda *args, **kwargs: ScaleRegularizeTrainerWrapper(modes[mode], *args, **kwargs)
    if trainable_camera:
        trainer = constructor(
            gaussians,
            scene_extent=dataset.scene_extent(),
            dataset=dataset,
            **configs
        )
    else:
        trainer = constructor(
            gaussians,
            scene_extent=dataset.scene_extent(),
            **configs
        )
    return trainer



from .densifier import *
from .opacity_reset import PartialOpacityResetter, PartialOpacityResetTrainerWrapper
from .combinations import LapisTrainer, DepthLapisTrainer, LapisCameraTrainer, DepthLapisCameraTrainer
from .combinations import Trainer, CameraTrainer, DepthTrainer, DepthCameraTrainer



import torch
import os
from gaussian_splatting.prepare import prepare_gaussians
from gaussian_splatting.render import rendering
from lapisgs.prepare import prepare_dataset


def prepare_rendering(sh_degree: int, source: str, device: str, trainable_camera: bool = False, load_ply: str = None, load_camera: str = None, load_depth=False, rescale_factor=1.0):
    dataset = prepare_dataset(source=source, device=device, trainable_camera=trainable_camera, load_camera=load_camera, load_depth=load_depth, rescale_factor=rescale_factor)
    gaussians = prepare_gaussians(sh_degree=sh_degree, source=source, device=device, trainable_camera=trainable_camera, load_ply=load_ply)
    return dataset, gaussians


if __name__ == "__main__":
    from argparse import ArgumentParser
    parser = ArgumentParser()
    parser.add_argument("--sh_degree", default=3, type=int)
    parser.add_argument("-s", "--source", required=True, type=str)
    parser.add_argument("-d", "--destination", required=True, type=str)
    parser.add_argument("-i", "--iteration", required=True, type=int)
    parser.add_argument("--load_camera", default=None, type=str)
    parser.add_argument("--rescale_factor", default=1.0, type=float)
    parser.add_argument("--mode", choices=["base", "camera"], default="base")
    parser.add_argument("--device", default="cuda", type=str)
    parser.add_argument("--no_rescale_depth_gt", action="store_true")
    parser.add_argument("--save_depth_pcd", action="store_true")
    args = parser.parse_args()
    load_ply = os.path.join(args.destination, "point_cloud", "iteration_" + str(args.iteration), "point_cloud.ply")
    save = os.path.join(args.destination, "ours_{}".format(args.iteration))
    with torch.no_grad():
        dataset, gaussians = prepare_rendering(
            sh_degree=args.sh_degree, source=args.source, device=args.device, trainable_camera=args.mode == "camera",
            load_ply=load_ply, load_camera=args.load_camera, load_depth=True, rescale_factor=args.rescale_factor)
        rendering(dataset, gaussians, save, save_pcd=args.save_depth_pcd, rescale_depth_gt=not args.no_rescale_depth_gt)
